
needs tutorial-markup

\<b>
***************************  GtkWindow Tutorial  ******************************

\<n>
A GTK Window is the toplevel widget of a GTK+ Widget hierarchy. Two window 
types are supported.
 
Windows of the type GTK_WINDOW_TOPLEVEL are managed by the window manager and
have a frame with a titlebar by default.
 
Windows of type GTK_WINDOW_POPUP are not managed by the window manager and do
not have a frame.
 
Nearly all windows you create should be of type GTW_WINDOW_TOPLEVEL.

\<?>
\<b> 
Creating a GTK+ Window

\<n>
To create a GTK+ Window we first have to load the module GtkWindow with

\<c>
  needs GtkWindow

\<n>
This implicitly loads all other modules needed by the class GtkWindow,
especially the oop and the gtk api modules.

\<?>
\<n>
Next we specify that the following definitions should go into the vocabulary
oop and we place the vocabulary gtk api on the top of the search order.

\<c>
  gtk api also  oop definitions

\<n>
Before going further let's have a look at the current cspForth status.

\<c>   ??

cr \<?>
\<n>
Now we will create our first window - named window - as an instance/object of
the class GtkWindow.

\<c>
  GtkWindow new window

\<n>
initialize it as a toplevel window

\<c>
  GTK_WINDOW_TOPLEVEL window init

\<n>
and make it appear on the display.

\<c>
  window show

\<n>
You should now see the new window on the desktop. It was made the active one.
Do not click on the windows close button because this would close the window.
 
 >>>>  Please click on this terminal window to redirect the input to it.  <<<<

\<?>

\<n>
Again, lets have a look on the cspForth system status.

\<c>   ??

cr \<?>

\<n>
The new window was created as the top word in the oop vocabulary. It was
created with a default size, a default title and a default icon. All this
can be changed.

\<?>
\<b>
Changing the Window Title

\<n>
The title is a window property and can be changed with the ! (store) method.

\<c>
  " GtkWindow Tutorial window" window title !

\<n>
The window now has the new title but it's to small, to fully show it. We can
change its size with the resize method.

\<?>
\<b>
Changing the Window Size

\<n>
The resize method expects to find the width and the heights (in pixels) on the
stack.

\<c>
  350 200 window resize

\<n>
Now you should be able to fully see the window title.

\<?>
\<n>
There are more size related window properties :
 
  default-size  default-height  default-width
  size-request  height-request  width-request   and   resizable
 
The default-xxx properties specify the default size to be used when a window
is created, the xxx-request properties specify the minimal window dimensions
and the boolean property resizable specifies if a window can be resized or not.

\<?>
\<b>
Changing the Window Icon

\<n>
To change the windows icon, we can use an icon from the GTK Stock, from a
GdkPixbuf or from a file. The most simple way is to take an icon from stock.
We can do it by assigning the stock-id of the icon to the icon-name property
of the window e.g.

\<c>
  " gtk-preferences" window icon-name !

\<n>
Now you should see the new icon in the windows title bar.
 
Note: You can browse the GTK Stock by means of the GdkPixbuf example.2

\<?>
\<n>
To take an icon from a file needs some more steps. We first have to load the
class GdkPixbuf

\<c>
  needs GdkPixbuf

\<n>
next we have to create a new GdkPixbuf as an invisible container for the icon
to be loaded from file

\<c>
  GdkPixbuf new pixbuf

\<n>
then we have to load the image from the file into the pixbuf

\<c>
  " icons/butterfly.png" /?/ pixbuf from-file init

\<n>
and finally we have to assign it to the windows icon property.

\<c>
   pixbuf @ window icon !

\<n>
The icon should have been changed again. It was sized automatically.

\<?>
\<n>
Wait, one more step is needed, we have to release the pixbuf data with

\<c>
  pixbuf free

\<b>
 
Hiding/Showing a Window 

\<n>
We can hide an existing window (it's not destroyed) with the hide method

\<c>
  window hide

\<n>
and we can make it visible again with the show method.

\<c>
  window show

\<n>
The visibility of a widget is controled by the boolean property visible so that
 
      window visible yes/no  and  true/false window visible !
 
would have the same effect as show and hide.

\<?>
\<b>
Closing a Window

\<n>
When the close button at the windows title bar is clicked, the window manager
sends a delete-event signal to the window and a signal handler is executed in
the background, that sends a destroy signal to the window which triggers
another signal handler that destroys the window.
 
We can bind our own signal handlers to a window, to add a different or
additional behavior.

\<?>
\<b>
Creating a delete-event signal handler 

\<n>
A delete-event handler is called with three parameters on the stack, the widget
identifier of the window that received the delete-event signal, an event-id
and a pointer to user defined data, that may be zero.
 
A delete-event handler must return a true or false flag. No destroy signal will
be send to a window, if the handler returns true.

\<?>
\<n>
Here we create our own delete-event handler in three steps. We first define a
headerless word, that prints a message and the data stack and returns a true
flag. Next we assign its execution token to a signal handler and finally we
bind this signal handler to the window.
 
Note: The value of the flag is hold in a variable, so that it can be changed.

\<c>
  variable flag  true flag !
 
  :noname ( wid event a|0 -- f )
          cr ." delete-event handler " .s reset flag @ ;
 
  3 20 cb on-delete-event
 
  " delete-event" on-delete-event 0 window signal connect drop
 

\<n>
Now, when clicking on the windows close button, the window is not closed but a
a message is displayed showing the data stack of the signal handler.

\<?>
\<n>
The handler was created by the word cb. We can display its glossary entry with

\<c>
  h cb

\<?>
\<b>
Creating a destroy signal handler

\<n>
Now we will also create a signal handler for the destroy signal. The handler is
called with two arguments, the widget identifier of the window that received
the destroy signal and a pointer - that may be zero - to user defined data. No
return value is expected.

\<c>
  :noname ( wid a|0 -- ) cr ." destroy handler " .s reset ;
 
  2 20 cb on-destroy
 
  " destroy" on-destroy 0 window signal connect drop

\<n>
Now, when clicking on the windows close button, the window is not closed and
the destroy handler is not executed, because there is still no destroy signal
emitted by our delete-event handler.

\<?>
\<n>
Remember that the flag, that is returned by our delete-event handler determines
whether a destroy signal is created or not. When this flag is false, a destroy
signal will be emitted. So we how have to change our flag variable accordingly.

\<c>
  false flag !

\<n>
Now click on the windows close button and the window will be destroyed. Notice
that the delete-event handler and the destroy handler are called.

\<?>
\<n>
Again let's look at the cspForth status.

\<c>   ??

cr \<n>
Notice, that the window object still exists in the vocabulary oop. Only the GTK
widget, that was assigned to it by the init method, was destroyed.

\<?>
\<n>
A GtkWindow has much more properties and methods as we have seen here. Please
see the libgtk2.0-doc for further information. 

\<?>
\<n>
The cspForth GTK+ API only implements a subset of all this methods and
properties. If you need more, it should not be to complicated, to implement
the needed ones too.

\<b> End of Tutorial

\<?>
normal bye


