/* libfloats.c       Copyright (C) 2007 Manfred Mahlow

Compiling the library without functions from glib :

  gcc -fPIC -Wall -g -c libfloats.c    creates the object file libfloats.o

  gcc -g -shared -Wl,-soname,libfloats.so.0 -o libfloats.so.0.0 libfloats.c -lc

  creates the library file libfloats.so.0.0

  See also: Program Library HOWTO

Compiling the library with functions from glib :

gcc -g -shared -Wl,-soname,libfloats.so.0 `pkg-config --cflags --libs glib-2.0` libfloats.c -o libfloats.so.0.0

Creates the file libfloats.so.0.0 in one run.

*/

#include <stdio.h>
#include <stdlib.h>
#include <glib-2.0/glib.h>


/* MM-080320

 Convert an ascii string to a float ( 32 bit IEEE floating point number ). 

   function declaration in forth :  __ascii2sf ( sp -- 0|c )

   function use in forth         :  __ascii2sf ( a sp -- a.sf 0|c )

*/
int __ascii2sf(int *sp)
{
   char *eptr;
   float f;

   f = (float) g_ascii_strtod((char *) sp[0], &eptr); 
   *sp = (int) &f;
   return((int) *eptr);
}




/*

 Convert a string to a double ( 64 bit IEEE floating point number ),
 with strtod from stdlib, using locales. Decimal point is then ',' in
 the locale de but '.' in the default locale. That means that the char
 for the decimal point changes from '.' to ',' when the gtk api is
 loaded in gsp4th. This is not ok for terminal input. Then the function
 g_ascii_strtod from glib should be used. See __ascii2df .

   import :  __str2df ( sp -- 0|c )
   use    :  __str2df ( a a sp -- dl dh 0|c )

   : >df ( a u -- df ) zt dup sp@ __str2df ;

*/
int __str2df(int *sp)
{
   char *eptr;
   double d;

   d = (double) strtod((char *) *sp, &eptr);
   *sp++ = ((int *)&d)[1];
   *sp   = ((int *)&d)[0];
   return((int) *eptr);
}


/*

 Convert an ascii string to a double ( 64 bit IEEE floating point number ).

   import :  __ascii2df ( sp -- 0|c )
   use    :  __ascii2df ( a a sp -- dl dh 0|c )

   : ascii>df ( a u -- df 0|c ) zt dup sp@ __ascii2df ;

*/
int __ascii2df(int *sp)
{
   char *eptr;
   double d;

   d = (double) g_ascii_strtod((char *) *sp, &eptr);
   *sp++ = ((int *)&d)[1];
   *sp   = ((int *)&d)[0];
   return((int) *eptr);
}



/*
Convert single to double ( 32 bit IEEE to 64 bit IEEE )

   import :  __s2df ( sp -- )
   use    :  __s2df ( sf sf sp -- dl dh )

   : s2df ( sf -- df ) dup sp@ __s2df ;
*/
void __s2df(int *sp)
{
    double d = *sp;
    *sp++ = ((int *)&d)[1];
    *sp   = ((int *)&d)[0];
}


/*
Convert double to int ( 64 bit IEEE to single )

   import :  __df2s ( dl dl -- n )
   use    :  __df2s ( df -- n )

   : df>s ( df -- n ) __df2s ;
*/
int __df2s(double d)
{
   return((int) d);
}


/*
Convert double to float ( 64 bit to 32 bit IEEE )

   import :  __df2sf ( dl dl -- sf )
   use    :  __df2sf ( df -- sf )

   : df>sf ( df -- sf ) __df2sf ;
*/
float __df2sf(double d)
{
   return((float) d);
}


/*
Convert float to double ( 32 bit to 64 bit IEEE )

   import :  __sf2df ( sp sf -- dh )
   use    :  __sf2df ( sf sp sf -- df )

   : df>sf ( sf -- df ) sp@ over __df2sf ;
*/
int __sf2df(int *sp, float f)
{
  double d = (double) f;
    *sp  = ((int *)&d)[0];
    return(((int *)&d)[1]);
}






