\  libs.4th
\ -----------------------------------------------------------------------------
\  iface libs , class lib          OOP Module for csp4th             MM-060901
\ -----------------------------------------------------------------------------
\
\                  Copyright (C) 2008 Manfred Mahlow
\
\ This is Open Source Software released under the "Artistic License". See the
\ enclosed LICENSE file.
\ -----------------------------------------------------------------------------
 
  needs Cell
  needs String
 
  oop definitions
 
  iface libs
  class lib
 
  libs definitions  self seal
 
  Cell new last   0 self last init
 
  lib definitions
 
  private  Cell   this hdl       \ holds the library handle
  public   String this name      \ holds the library name
  private  Cell   this link      \ holds the oid of the next older lib
           Cell   this f.last    \ holds the oid of the last imported function

  self seal
 
  public
 
  m: open (O -- )
     this hdl @ if exit then \? cr ." opening " this name ?
     this name @ (lib) dup this hdl !
     0if ." Error : could not open " this name ? ."   " -1 throw then
   ;
 
  m: find (O a u -- hdl )
     2dup this hdl @ dup 
     if (func) 
     else drop cr type ."  ? " this name ? ."  is closed  " -1 throw
     then dup 
     0if drop cr type ."  ? Not found in " this name ? ."   " -1 throw then
     nip nip
   ;
 
  m: init (O a u -- )
     128 chars this name init  this name !  \ MM-080422
     0 this f.last init
     libs last @ this link init  this oid libs last !  this open
   ;
 
 
m: f.link ( a.f -- a.f.link ) ;
m: f.hdl ( a.f --  a.f.hdl ) cell+ ;
m: f.#o ( a.f -- a.f.#o ) cell+ cell+ ;
m: f.#i ( a.f -- a.f.#i ) cell+ cell+ cell+ ;
m: f.oid.lib ( a.f -- a.f.oid.lib ) cell+ cell+ cell+ cell+ ;
m: f.name ( a.f -- a.wa ) cell+ cell+ cell+ cell+ cell+ ;
 
0 [IF] \ for debugging only
 
m: f? ( f.oid -- )
       cr ."     f.oid = " dup .
       cr ."    f.link = " dup self f.link @ .
       cr ."     f.hdl = " dup self f.hdl @ .
       cr ."      f.#o = " dup self f.#o @ .
       cr ."      f.#i = " dup self f.#i @ .
       cr ." f.oid.lib = " dup self f.oid.lib @ dup . 
          ." hdl = " dup self hdl ? ." name = '" self name ? ." '"
       cr ."    f.name = " self f.name @ dup . ." name = '" count type ." '"
       cr cr ;
[THEN]
 
m: #err ( -- ) ." ?  Faulty stack diagram. " -1 throw ;
 
m: #i ( max.#i -- #i )
 \ Parse stack diagram for number of input parameters.
   0 do parsews " --" cmp 0if i unloop exit then loop self #err ;
 
m: #o ( max.#o -- #o )
 \ Parse stack diagram for number of output parameters.
   0 do parsews " )" cmp 0if i unloop exit then loop self #err ;
 
m: f.call (S f.oid -- )
          (O -- )
 \ Call the function object f.oid . If the functions handle is zero reload
 \ the function from its library. Open the library if it is closed.
   this f.hdl @ dup
   0if  \ reopen library and get the function handle
       drop
       this f.oid.lib @ self open    \ reopen library
       this f.name @ count ""        \ get function name zero terminated
       \?  cr ." reloading " 2dup type space \ print name
       this f.oid.lib @ self find    \ get function handle
       dup this f.hdl !              \ and save it
   then
   this f.#i @ (call)                \ call function
   this f.#o @ 0if drop then ;       \ drop garbage if void function
 
  m: ( (O "i*x -- j*x" a1 u1 a2 u2 -- )
     this open  \? cr ." importing " 2over type
     2over "" this find over >r >r 
     dup if (? ."  as " 2dup type ) else 2drop 2dup then created
     here this f.last @ , this f.last !
     r> ,     \ f.hdl
     15 self #i 15 self #o  \ parse the stack diagram
     \? ."  ( " over . ." -- " dup . ." ) from "  this name ? ."   "
     ,         \ f.#o
     ,         \ f.#i
     this oid ,  \ f.oid.lib
     r> 0if last @ >name , 2drop 
      \ else here cell+ , here over 2 + allot place    \ MM-090304
        else here cell+ , here over 2 + allot placez
        then \ f.name
     does> ( -- a ) self f.call
   ;
 
  : import (S "name" oid -- a u 0 0 oid )
    >r parsews 0 0 r>
  ;
 
  : as (S "name" a1 u1 0 0 oid -- a1 u1 a2 u2 oid )
    >r 2drop parsews r>
  ;
 
  hide #err  hide #i  hide #o
 
  m: ? (O -- )
     this name ? ."  :  "  \ MM-071229
     this f.last @ ?dup 0if ." No functions/words imported. " exit then
     begin dup self f.name @ count type space \ ."  : "  
            dup self f.hdl @ . space space @ dup 0until drop
   ;
 
  m: reset (O -- )
     this f.last @ 0; begin dup self f.hdl off @ dup 0until drop
   ;
 
  m: close (O -- )
     this hdl @ dup 0if drop exit then
     (-lib) 0 this hdl ! ( crash !? ) this reset ;
 
  hide f.last  hide f.call  hide f.name  hide f.oid.lib  hide f.#i
  hide f.#o  hide f.hdl   hide f.link
 
  libs definitions
 
  m: ? (S -- )
     self last @ dup 0if drop ." No shared libraries used." exit then
     begin cr dup lib name ? ."  : " 
               dup lib hdl @ . 
                   lib link @ dup 0until drop ;
 
  m: reset (S -- )
     self last @ 0;
     begin dup lib reset lib link @ dup 0until drop ;
 
  m: reopen (S -- )
     self last @ 0;  \ exit if no libs loaded
     0 swap begin dup lib link @ dup 0until drop  \ get oids of the libs
     begin lib open dup 0until drop ;
 
  m: close (S -- )
     self last @ 0; begin dup lib close lib link @ dup 0until drop ;
 
  self ' close onexit  \ execute on exit
 
  oop definitions 
\ ----------------------------------------------------------------------------
\  Last revision: MM-090307  DO LOOP -> do loop ...
\                 MM-090304  place -> placez
\                 MM-090101  do loop ... -> DO LOOP ...
\                 MM-080920  (create) -> created
\                 MM-080531  Reva repeat while -> begin 0until
\                 MM-080422  string size for lib init changed. Name should
\                            better be allocated.

