\  oop-mk2.4th
\ -----------------------------------------------------------------------------
\  Classes, Objects and Methods for csp4th                           MM-070222
\ -----------------------------------------------------------------------------
\ 
\                     Copyright (C) 2008 Manfred Mahlow
\ 
\  This is Open Source Software released under the "Artistic License". See the
\  enclosed LICENSE file.
\ -----------------------------------------------------------------------------
 
  needs search-order
 
  only definitions  forth also  hidden also   decimal
 
  macro
  : .. (C -> .. ) (S -- )  \ compiling , interpreting
    (esm) off (osm) off ;
 
  : definitions (S -- )
    definitions (private) off  (esm) @  (ecm) ! ;
 
  ' ??  alias ??? (S -- )

\ ----------------------------------------------------------------------------- 
  hidden definitions
\ -----------------------------------------------------------------------------
 
  variable c.link   0 c.link !

  variable (union)  (union) off

  \  The body of a class :
 
  \  +------------------------+
  \  |        wa(last)        |  addr = cid = context|current @
  \  +------------------------+
  \  |        wa(self)        |  *)
  \  +------------------------+
  \  | cid(previous context)  |
  \  +------------------------+
  \  | cid(container context) |
  \  +------------------------+
  \  | addr units per object  |
  \  +------------------------+
  \  | cid(inherited context) |
  \  +------------------------+
 
  \ *) bit 0 of wa is used as lock flag,
  \    bit 1 of wa is used as private flag
 
  : c.self ( cid -- a ) cell+ ;
 
  : c.container ( cid -- a ) [ 3 cells ] literal + ;
 
  : o.units  ( cid -- a )    [ 4 cells ] literal + ;
 
  : c.inherited ( cid -- a ) [ 5 cells ] literal + ; 
 
  : c.xt ( cid -- xt ) c.self @ [ %11 invert ] literal and >xt @ ;
 
  : c?defs ( -- )
    (ecm) @ 0if ." Error: (non-default) compile mode missing  " -1 throw then
  ;
 
  : c?objects ( cid -- )
      o.units @ 0if .cn ." Error: no object defined   " -1 throw then ;
 
  : o.seal ( cid -- )
      c.self dup @ %1 or swap ! ;
 
  : o?sealed ( cid -- )
      c.self @ %1 and 0; ." Error: Class is sealed   " -1 throw ;

  : o?union ( u1 u2 -- u3 ) \ if (union) = true
            ( u -- )        \ if (union) = false
            (union) @ if max else current @ o.units +! then ;

  : class-wordlist (S "name" -- cid )
    create here                        \ cid of the new class
    0 ,                                \ set empty wordlist
    current @ @ ,                      \ wa of the new class
    c.link @ ,  c.link !               \ add class to c.link
    (ecm) @ if current @ else 0 then , \ store container class
    0 ,                                \ no object defined (no address units)
    0 ,                                \ cid of the inherited class (none)
  ;

  class-wordlist c.root
  class-wordlist c.methods
  class-wordlist c.macros
  class-wordlist o.methods
  class-wordlist o.macros

  variable (xt)

  : foreach-class ( xt -- )
   \ executes xt ( cid -- f ) for every class of the current  class search
   \ order. Exits if xt returns a true flag.
     (xt) !
     e.context begin @ ?dup while
       dup >r (xt) @execute if r> drop exit then r> c.inherited
     repeat
     e.context @ o.units @ if \ class search order
       compiling? if
         o.macros (xt) @execute if exit then
         (osm) @ 0if c.macros (xt) @execute if exit then then
       else
         o.methods (xt) @execute if exit then
         (osm) @ 0if c.methods (xt) @execute if exit then then
       then
     else \ abstract class (interface) search order
       (osm) @ 0if
         compiling? if c.macros else c.methods then (xt) @execute if exit then
       then
     then
     c.root (xt) @execute drop
   ;

   hide (xt)

  :noname ( a u 0 cid -- wa wa )
          ( a u 0 cid -- a u 0 0 )
     nip (c.find) dup ;

  :noname ( a u -- wa )     \ word found
          ( a u -- a u 0 )  \ word not found
     (esm) @ 0if [ ' (find) defer@ ] compile, exit then
     0 [ swap ] literal foreach-class
  ; is (find)

\ -----------------------------------------------------------------------------
  forth definitions
\ -----------------------------------------------------------------------------

  : class (S "name" -- )
      [compile] macro class-wordlist
      does> ( cid -- )
            (osm) @ 0if prelude? if (osm) on else (osm) off then then
            e.context ! (esm) on ;

  ' class alias iface (S "name" -- )
 
  : .classes (S -- )
    c.link @ .clink ;
 
  : method (S -- )
    ['] .. prelude ;
 
  : m' (S "name" -- xt )
    ' method ;
 
  defer m: (S "name" -- )
        :noname ( -- ) method : ; is m:
 
  macro
  : self (C -> current ) (S -- )  \ compiling , interpreting
    current @  (ecm) @ if  (esm) on (osm) off then context ! ;
 
  :noname (S a u cid -- wa )     \ word found
          (S a u cid -- a u 0 )  \ word not found
     [ ' (c.find) defer@ ] compile, dup 0if ( a u 0 -- ) exit then ( wa -- )
     (osm) @ 0if exit then ( wa -- )
     dup >class @ %10 and if ( private ) >name count false then
  ; is (c.find)
 
  : private (S -- )
    c?defs %10 (private) ! ;
 
  : public (S -- )
    c?defs (private) off ;
 
  : cb ( "name" xt numparams stacksize -- ) 
       (ecm) @ if private method cb public else cb then ;

  : stdcb ( "name" xt numparams stacksize --   ) method stdcb ;

  : show-all ( cid -- 0 )
      cr 80 0 do '- emit loop show-wordlist 0 ;

\ -----------------------------------------------------------------------------
  c.root current ! (ecm) on  \ = c.root definitions
\ -----------------------------------------------------------------------------
 
  macro
  : self (C -> current ) (S -- )
    [compile] .. [compile] self ;
 
  macro
  ' (  alias ( (S -- )  \ Alias of the core word.

  macro
  ' [  alias [ (S -- )  \ Alias of the core word.
 
  ' ]  alias ] (S -- )  \ Alias of the core word.

  alias: \ | (S -- )
 
  macro
  ' ..  alias .. (S -- xt )  \ Alias of the word .. in the root vocabulary.
 
  ' .s  alias .s (S -- )     \ Alias of the stack printing word.
 
  ' words  alias words (S -- )  \ Alias of the core word.


  :noname ( cid -- 0 ) .cn space 0 ;

  m: ?s ( -- )
    cr ."   " .s
    cr ." Current: " current @ .cn
    cr ." Context: " literal foreach-class ;

  : ?? ( -- ) \ Display the current system status
    self ?s  e.context @ show-all drop ;

  : ??? (S -- )
      self ?s  ['] show-all foreach-class ;

  hide ?s

  macro
  : ; (S -- ) (C <iface> -> .. )
    compiling? 0if ." ; ?  compile mode missing  " -1 throw then
    last @ 0if ." ; ?  default search order missing  " -1 throw then
    [compile] ;
    e.context @ c.xt 1 or last @ >class
    dup >r @ %10 and or ( private ? )
        r> !  [compile] .. ;
 
  forth [DEFINED] h [IF] ' h alias h ( "name" -- ) [THEN]

\ -----------------------------------------------------------------------------
  c.methods current !  (ecm) on  \ = c.methods definitions
\ -----------------------------------------------------------------------------

  m: definitions (S -- )  \ interpreting
       (esm) on  definitions  (esm) off  (union) off ;
 
  : ' (S "name" -- xt ) (C <iface> -> .. )
      ' [compile] .. ;
 
  : m' (S "name" -- xt ) (C <iface> -> .. ) 
      ' method [compile] .. ;

  : is (S "name" xt -- ) (C <iface> -> .. )
       is [compile] .. ;
 
  : [DEFINED] (S "name" -- flag ) (C <iface> -> .. )
    [compile] [DEFINED]  [compile] .. ;
 
  ' reveal  alias reveal (S "name" -- ) \ Alias of the core word.
 
   private
  m: align (S -- )  \ interpreting
     c?defs  current @  dup o?sealed  o.units @  \ cr  dup ." old size : " .
     cell+ 1 -  1 cells /  cells                 \ cr  dup ." new size : " .
     current @ o.units ! ;
 
  private
  m: allot (S u -- )  \ interpreting
     c?defs current @ o?sealed o?union ;
 
  private
  m: seal (S -- )  \ interpreting
     c?defs current @ dup o?sealed o.seal ;
 
  private
  m: units (S -- u )  \ interpreting
     e.context @ o.units @ ;
 
  private
  : inherit (S -- ) (C <class> -> .. )
    c?defs  e.context @  current @  dup o?sealed
    2dup = if drop .cn ." inherit ?  Can not inherit myself  " -1 throw then
    dup c.inherited @  over o.units @ or
    if over .cn ." inherit ?  Can not inherit any more  " -1 throw then
    over o.units @ over o.units !   \ inherit the object size
    2dup c.inherited !              \ inherit the class context
    drop o.seal [compile] .. ;      \ seal the inherited class
 
  private
  : super (C <class.1> -> <class.2> ) (S -- )
    e.context @ c.inherited @ 
    dup 0if ." super ?  No inherited class  " -1 throw then e.context ! ;
 
  macro
  : up (S -- ) (C <iface1> -> <iface2> )
    e.context @  c.container @
    dup 0if ." up ? no container context  " -1 throw then  e.context ! ;

\ -----------------------------------------------------------------------------
  o.methods current ! (ecm) on  \ = o.methods definitions
\ -----------------------------------------------------------------------------
 
  private
  : new (S "name" -- ) (C <class> --> .. )
    e.context @  dup c?objects  dup c.xt prelude  create
    dup o.units @ 0  do 0 c, loop  o.seal  [compile] .. ;

  private  \ MM-081004
  : this (S "name" -- ) (C <class> -> .. )
    c?defs  current @ o?sealed  e.context @ dup >r c?objects
    r@ c.xt prelude  create  current @ o.units @ ,
    r@ o.units @  o?union r> o.seal  [compile] ..
    does> ( oid a.offset -- oid' ) @ + ;
 
  public
  m: oid (S oid -- oid )
   ;

\ -----------------------------------------------------------------------------
  c.macros current ! (ecm) on  \ = c.macros definitions
\ -----------------------------------------------------------------------------
 
  private macro  c.methods e.context ! (esm) on
  ' super  alias super (C <class.1> -> <class.2> ) (S -- )
 
  private macro 
  : units (C <class> -> .. ) (S -- )  \ compiling
          (S -- u )                   \ executing
    [compile] ..  [ c.methods e.context ! (esm) on ] units  [compile] literal ;
 
  macro  c.methods e.context ! (esm) on
  ' up  alias up (S -- ) (C <iface1> -> <iface2> )
 
  macro
  : ['] (S "name" -- ) (C <iface> -> .. )
    [compile] ['] [compile] .. ;
 
\ -----------------------------------------------------------------------------
  o.macros current ! (ecm) on  \ = o.macros definitions
\ -----------------------------------------------------------------------------
 
  public
  m: oid (S oid -- oid )
   ;
 
1 [IF]  \ Use object stack mechanism ? Yes !

\ -----------------------------------------------------------------------------
  hidden ( object stack ) definitions
\ -----------------------------------------------------------------------------
 
  variable (-os)  (-os) on   \ ON if the method 'this' is not supported in a
                             \ methods definition.
 
  variable (oid)  (oid) off  \ Top of the object stack (TOOS). The other 
                             \ values of the object stack are hold on the
                             \ return stack.
 
  : (this) ( -- oid )
  \ Return the TOOS on the data stack.
    (oid) @ ;
 
  : o.pop ( -- )
  \ Drop the TOOS
    r> (oid) ! ;
 
  : o.push ( oid -- )
  \ Push the oid from the TOS to TOOS.
    r> (oid) @ >r ['] o.pop >r >r (oid) ! ;


  c.root current ! (ecm) on  \ = c.root definitions
 
  private
  ' dup alias dup 
  \ Needed to hand over the oid of the calling object to methods of the
  \ called one.
  \ Usage:  this ( -- oid.1 ) dup <object> ( -- oid.1 oid.2 ) <method>
  \         ( -- oid.1 ) self dup <object> ( -- oid.1 oid.2 ) <method> 

\ ----------------------------------------------------------------------------- 
  forth definitions
\ -----------------------------------------------------------------------------

  loc:

   : opush ( -- ) \ push TOOS to the return stack      MM-080324
      r> (oid) @ >r >r ;

   : opop ( -- ) \ pop TOOS from the return stack     MM-080324
      r> r> (oid) ! >r ;

   macro
   : catch ( xt -- n )
     \ The top of the object stack must be saved and restored when the 
     \ object stack is used in a method.
     compiling? 0if ." catch ? compile mode missing  "  -1 throw then
     (ecm) @ if ['] opush [compile] compile, then
     ['] catch [compile] compile,
     (ecm) @ if ['] opop [compile] compile, then ;

   reveal catch

  loc;

  :noname (S "name" --)
     [ ' m: defer@ ] compile,  here (-os) !
   ; is m: ( "name" --)

  : :noname (S -- xt )
      :noname here (-os) ! ;

  : : (S "name" -- )
    : here (-os) ! ;
 
  macro
  : this (C -> <class> ) (S -- )  \ compiling
         (S -- oid )              \ executing
    c?defs  compiling? 0if ." this ? Compile mode missing.  " -1 throw then  
    (-os) @ if ." this ?  Object stack underflow.  " -1 throw then
    [compile] self  ['] (this) [compile] compile, ;
 
  macro
  : (O (S -- ) \ compiling
       (S oid -- ) \ executing , moves TOS to TOOS
    c?defs  compiling? 0if ." Error: Compile mode missing  " -1 throw then
    (-os) @ here = 0if ." (O ? invalid use of the object stack  " -1 throw then
    ['] o.push  [compile] compile,  (-os) off  [compile] ( ;
 
[THEN]

  : union{ ( -- 0 )  \ MM-081004
      c?defs  0 (union) on ;

  : }union ( u -- )  \ MM-081004
      c?defs current @ o.units +! (union) off ;
 
\ -----------------------------------------------------------------------------
  hidden definitions
\ -----------------------------------------------------------------------------

  hide o?union  hide (union)
 
\ -----------------------------------------------------------------------------
  forth definitions
\ -----------------------------------------------------------------------------

  hide show-all
 
  verbose @ not [IF] hide (private)  hide (prelude) [THEN]

  vocabulary oop (S -- )
 
  only  forth also  oop definitions also
 
\ -----------------------------------------------------------------------------
\  Last revision : MM-071227
\                  c.create removed
\                  MM-071226
\                  'c.xt execute' replaced by 'c.set-order'
\                  MM-071223
\                  class modified to make iface.src obsolet
\                  A class without data ( units = 0 ) is an interface or
\                  abstract class. It can inherit from another interface or
\                  class as long as it is not sealed.
\                  MM-080108
\                  h added to c.root
\                  MM-080208
\                  MM-080303  cb changed to create private callback handlers in
\                             class and iface contexts
\                  MM-080324  catch redefined to save and restore the TOOS
\                  MM-080409  stack diagrams
\                  MM-080420  \ added to c.root
\                  MM-080530  Reva repeat while -> begin 0until
\                  MM-081004  union{ and }union added
\                  MM-090101  do loop ... -> DO LOOP ...
\                  MM-090117  method ['] added
\                  MM-090306  :: -> :noname
\                  MM-090307  DO LOOP -> do loop ...
\                  MM-090509  (find) changed to directly search linked 
\                             lists of classes in the extra search mode
\                  MM-090510  class changed not to create a class search order
\                             but only set the e.context.
\                             ??? up super self this modified accordingly
\                             class-wordlist added , c.root , c.methods , 
\                             c.macros , o.methods and o.macros are now
\                             class-wordlists
\                             (c.root) ... removed
\                             Does not need the search-order extension any more.
