\  search-order.4th
\ ------------------------------------------------------------------------------
\                    Search order extension for csp4th
\ ------------------------------------------------------------------------------
\
\                    Copyright (C) 2008 Manfred Mahlow
\
\ This is Open Source Software released under the "Artistic License". See the
\ enclosed LICENSE file.
\ ------------------------------------------------------------------------------

  needs help

\ wa = word address, the address of a words header in the dictionary

\ cid = context identifier, points to the wordlist of a vocabulary

 decimal

 :noname (S a u cid -- wa )     \ word found
         (S a u cid -- a u 0 )  \ word not found
    \ Find the word named a,u in the wordlist, specified by cid. Return its
    \ dictionary address if found, return the string a,u and false, if not
    \ found.
    s.context ! [ ' (find) defer@ ] compile,
 ; defer! (c.find)

 :noname ( a u -- wa )     \ word found
         ( a u -- a u 0 )  \ word not found
    \ Find the word a,u in the current search order. Return its dictionary
    \ address if found, return the string a,u and false, if not found.
    context dup @ over cell+ @ =if cell+ then
    begin dup >r @ dup 0if r> drop exit then (c.find)
           dup if r> drop exit else drop then r> cell+ again
 ; is (find)


 variable (root)  current @ (root) ! \ to create root with a container field = 0
 variable (forth) 0 (forth) !
 variable v.link  0 v.link !         \ linked list of all vocabularies

 : vocabulary ( <name> -- )
      create  here                    \ cid of the new vocabulary
              0 ,                     \ initially empty wordlist
              current @ @ ,           \ wa of this vocabulary
              v.link @ , v.link !     \ cid of the last voc in the v.link
              current @ dup (root) @ = over (forth) @ = or
              if drop 0 then ,        \ store container, if not root or forth
              does> ( cid -- )
                v.context !  \ store as top item in the vocabulary search order
                (esm) off ;  \ disable the extra search mode

\ csp4th supports two search orders controlled by the variable (esm), the
\ vocabulary search order, (esm)=0, and an extra search order, (esm)<>0, 
\ which was implemented to be used by interfaces, classes and objects.


    \  The body of a vocabulary :

    \  +---------------------------+
    \  |          wa(last)         |  addr = cid = context|current @
    \  +---------------------------+
    \  |          wa(self)         |  bit 0 may be used as a flag
    \  +---------------------------+
    \  |    cid(prev-vocabulary)   |
    \  +---------------------------+
    \  | cid(container-vocabulary) |
    \  +---------------------------+

\ ------------------------------------------------------------------------------
  vocabulary root  root v.context @ dup (root) !  current !   \ root definitions
\ ------------------------------------------------------------------------------

 : get-order ( -- cid1 ... cidn n )
       0 context begin dup @ if cell+ swap 1+ swap dup else drop 0 then 0until
       dup >r begin 1- dup cells context + @ swap dup 0until drop r> ;


  : set-order ( cid1 ... cidn n -- )
        dup #vocs >if -49 throw then   \ -49 = search order overflow
        context swap begin 1- >r dup cell+ >r ! r> r> dup 0until swap ! ;

 : also ( -- )
   get-order over swap 1+ set-order ;  \ MM-071210
\  get-order >r 2dup = 0if dup r> 1+ else r> then set-order ;

 : previous ( -- )
       get-order nip 1- dup 0if -50 throw then set-order ;
       \ -50 = search order underflow

 : only [ current @ ] literal dup 2 set-order ;

 : definitions context @ current ! ;

 ' .s alias .s

  vocabulary forth

  forth context cell+ @ @  context @ !    \ link the core wordlist to the
                                          \ forth vocabulary

  context @ (forth) !

  root only forth also  

  root definitions

\ ------------------------------------------------------------------------------
  vocabulary hidden   hidden definitions also
\ ------------------------------------------------------------------------------

  ' v.link alias v.link

  : (.cn) ( cid -- )
  \ Print the name of the context 'cid'
    cell+ @ $FFFFFFFC and >name count type ;   \ clear flags

  : .cn ( cid -- )
  \ Print the name of the context cid with all its containers.
    0; dup %1 and 0if . exit then  \ print wordlist id and exit
       0 swap begin >r 1+ r@ #3 cells + @ dup 0until drop
              begin r> (.cn) space 1- dup 0until drop ;

  : .clink ( link -- )
         0; begin dup .cn space cell+ cell+ @ dup 0until drop ;

\ ------------------------------------------------------------------------------
  root definitions
\ ------------------------------------------------------------------------------

  : .vocs ( -- )
  \ List all existing vocabularies.
    v.link @ .clink ;

  : order ( -- )
    get-order begin swap .cn space 1- dup 0until drop ;

  : .context
  \ Display the current compilation context and the current search order.
    ." Current: " current @ .cn cr
    ." Context: " order ;

  : ?? ( -- )
  \ Display the current system status
    \ \? cr (osm) @ if ." > object search mode <" then
    cr ."   " .s cr .context cr
    ." ----------------------------------------"
    ." ----------------------------------------"
    cr words
  ;

  ' h alias h ( "name" -- )
              \ Display help for name.

  only forth also definitions

  hide (root)
  hide (forth)
  hide v.link  \ v.link is still visible in the vocabulary hidden
\  hide s.context

  : wordlist ( -- wid )
             \ Create a new empty wordlist and return its wordlist identifier
             \ wid on the stack.
             \ Note: wordlist identfiers are aligned (even) addresses. 
             \       context identfiers of vocabularies, ifaces and classes
             \       are unaligned (odd) addresses. This is used to identify
             \       wordlists, that need special treatment.
    align  here 0 , 
  ;
\ ------------------------------------------------------------------------------
\ Last revision: MM-090405  revaOE -> csp4th
\                MM-090306  :: -> :noname
\                MM-080530  Reva repeat while -> begin 0until
\                MM-080205
\                needs newhelp and h added
\                MM-080107
\                s.context made visible

