; Linux core:
include 'macros'
OS = 0

format PE console
stack 16000
entry _start

struc input_record {
	.type	rd 1
	.keydown rd 1 
	.repeatcount rw 1
	.vkeycode rw 1
	.vscancode rw 1
	.ascii rw 1
;	.unicode rw 1
	.controlkeystate rd 1
}

include "inc/win32a.inc"
; These macros must be defined prior to the general code:
macro NONBSS {
section '.data' data readable writeable 
align 4
rb 4
}
macro BSS { 
bss_start:
ir input_record
;processheap rd 1
emit_buffer rd 1
written_buffer rd 1
; fhandle rd 1
numbytes rd 1
x rd 1
consolemode rd 1
}

KEY_EVENT =1
MOUSE_EVENT =2
WINDOW_BUFFER_SIZE_EVENT =4
MENU_EVENT =8
FOCUS_EVENT =16
CTRL_C_EVENT = 0
CTRL_BREAK_EVENT = 1
CTRL_CLOSE_EVENT = 2
CTRL_LOGOFF_EVENT = 5
CTRL_SHUTDOWN_EVENT = 6
ENABLE_LINE_INPUT = 2
ENABLE_ECHO_INPUT = 4
ENABLE_PROCESSED_INPUT = 1
ENABLE_WINDOW_INPUT = 8
ENABLE_MOUSE_INPUT = 16
ENABLE_PROCESSED_OUTPUT = 1
ENABLE_WRAP_AT_EOL_OUTPUT = 2

section ".code" code readable writeable executable
;---------------------------------------------------------------------
; windows-specific linkage stuff:
align 4
os_start:
	invoke GetModuleHandle, 0
	mov [hinstance], eax
	push eax
	invoke GetModuleFileName, eax, app_file_name, 255
	pop eax

; iinvoke GetProcessHeap
;	invoke HeapCreate, 0, 0, 0 ; HEAP_NO_SERIALIZE
;	mov [processheap], eax

	invoke GetStdHandle, STD_INPUT_HANDLE
	mov [StdIn], eax

	invoke GetEnvironmentStrings
	mov [environ], eax

	invoke GetStdHandle, STD_OUTPUT_HANDLE
	mov [StdOut], eax

.redirected:

	invoke GetCommandLine
	mov [argv], eax
	push esi
	; figure out how many args there are, while modifying the argv if necessary:
	mov esi, eax
	mov edi, esi
	xor eax, eax
	mov [argc], eax		; initially, zero args

	; normal argument
.argloop:
	lodsb		; ESI, source string
	cmp al, '"'
	je .quoteloop
	cmp al, ' '	; space
	je .argspace
	cmp al, 9	; tab
	je .argspace
	or al, al
	jz .argend
	stosb
	jmp .argloop

.quoteloop:
	lodsb		; ESI, source string
	or al, al
	jz .argend
	cmp al, '"'	; space
	je .argspace
	stosb
	jmp .quoteloop

.argspace:	; esi points to a space character
	mov [edi], ah	; NUL terminate
	inc edi
	; scan to end of spaces
.argspace2:
	lodsb
	cmp al, ' '
	je .argspace2
	cmp al, 9
	je .argspace2
	or al, al
	jz .argend
	; now ESI points to next argument, and EDI points to next place to
	; store:
	dec esi
	inc [argc]
	jmp .argloop

.argend:
	inc [argc]
	mov [edi], ah

	pop esi
	ret

; 	; set up 'per thread' handler
; 	pop eax		; return address
; 	xor ebx, ebx
; 	push ebp
; 	push dword 0
; 	push dword 0
; 	push dword [fs:ebx]
; 	push criterr
; 	push dword [fs:ebx]
; 	mov dword [fs:ebx], esp
; 	jmp eax		; go back to caller

; Critical error handler:
; ALIGN 4
; criterr:
; 	mov  eax, [esp+4]	; esp+4 --> exception record
; 				; esp+8 --> CONTEXT record
; 				; esp+12 -->  ERR structure
; 				; EXCEPT+0 -> exception code
; 				;  + 4 -> flags
; 				;  + 8 -> nested exception
; 				;  + 12  -> exception address
; 				;  + 16 -> numberparams
; 				;  + 20 -> additional data
; 	mov ebx, [eax]
; 	mov [except_code], ebx	; 
; 	mov ebx, [eax+12]
; 	mov [except_addr], ebx
; 	call reset
; 	call exception
; 	mov eax, [esp+12]
; 	mov dword [eax+0xb8], hello
; 	mov eax, 0	; 1 = next handler, 0= continue
; 	ret


;---------------------------------------------------------------------

;---------------------------------------------------------------------
; IMPLEMENTATION OF OS-DEFERRED WORDS
;---------------------------------------------------------------------
PROC os_bye
	upop ebx
	invoke ExitProcess,ebx
ENDP os_bye
align 4
os_type:
	mov ebx, [esi]
	pusha
	invoke WriteFile, [StdOut], ebx, eax, written_buffer, 0
	popa
	drop2
	ret

align 4
os_emit:
	pusha
	mov [emit_buffer], eax  ; keep it safe
	invoke WriteFile, [StdOut], emit_buffer, 1, written_buffer, 0
	popa
	drop
	ret

align 4
os_key:
	dup
	push esi
	invoke ReadFile, [StdIn], emit_buffer, 1, written_buffer, 0
	pop esi
	mov eax, [emit_buffer]
	ret

; ( s n <name> -- )
PROC _loadlib
	call _zt
	push esi
	invoke LoadLibrary, eax
	pop esi
	ret
ENDP _loadlib
PROC _unloadlib
	test eax, eax
	jz .done
	push esi
	invoke FreeLibrary, eax
	pop esi
.done:
	drop
	ret
ENDP _unloadlib

; ( s n lib -- handle )
PROC _osfunc
	upop ebx
	lodsd
	push esi
	invoke GetProcAddress, ebx, eax
	pop esi
	ret
ENDP _osfunc
	
;---------------------------------------------------------------------
; ANS FILE ACCESS WORDS
;---------------------------------------------------------------------
; ( -- fam )
os_ro = GENERIC_READ
os_rw = GENERIC_READ OR GENERIC_WRITE
os_wo = GENERIC_WRITE

; ( c-addr u fam -- fileid ) \   s" file.txt" r/w create-file
PROC io_create
	mov ecx, CREATE_ALWAYS
	mov ebx, os_rw
	jmp openrw.2
ENDP io_create

PROC openr
	mov ebx, os_ro
	jmp openrw.1
ENDP openr

; ( c-addr u -- fileid ) \   s" file.txt" r/o open-file
PROC openrw
	mov ebx, os_rw
.1:
	mov ecx, OPEN_EXISTING

.2:
	mov [__ior], 0
	call _zt
	push esi
	invoke CreateFile, eax, ebx, FILE_SHARE_READ, 0, ecx, FILE_ATTRIBUTE_NORMAL, 0
	pop esi
	cmp eax, INVALID_HANDLE_VALUE
	jne .open.0
	push eax
	invoke GetLastError
	mov [__ior], eax
	pop eax
.open.0:
	mov ebx, [__ior]
	or ebx, ebx
	ret
ENDP openrw

; ( fileid -- ior )
PROC io_close
	push esi
	invoke CloseHandle, eax
	pop esi
	jmp win_return2
ENDP io_close

; ( c-addr u1 fileid -- u2 ) 
PROC io_read
	upop ecx ; [fhandle]; handle
	upop [numbytes]; how many
	push esi
	invoke ReadFile, ecx , eax, [numbytes], written_buffer, 0
	pop esi
	upsh [written_buffer]
	swap
	jmp win_return2
ENDP io_read

; ( c-addr u fileid -- ior )  pad 10 file_desc @ write-file
PROC io_write
	upop ecx ; [fhandle]; handle
	upop [numbytes]; how many
;	mov [tempX], eax ; buffer
	push esi
	invoke WriteFile, ecx, eax, [numbytes], written_buffer, 0
	pop esi
win_return2:
	or eax, eax
	jz .err
	xor eax, eax
.err2:	test eax, eax
	upop [__ior]
	ret
.err:	invoke GetLastError
	jmp .err2
ENDP io_write

;FORTH 'GetLastError', _gle
;	dup
;	invoke GetLastError
;NEXT _gle
; ( fileid -- ud )
PROC io_size
	push esi
	invoke GetFileSize, eax ,0 ; [fhandle], 0
	pop esi
	ret
ENDP io_size

; ( c-addr u -- x ior )
; PROC io_status
; 	call _zt
; 	upop ebx
; 	pusha
; 	invoke GetFileAttributes, ebx
; 	mov [tempX], eax
; 	popa
; 	upsh [tempX]
; 	xor ebx, ebx
; 	cmp [tempX], -1
; 	je .io.1
; 	inc ebx
; .io.1:
; 	upsh ebx
; 	ret
; ENDP io_status

; ( fileid -- ud ior )
PROC io_tell
	upop ebx
	mov [tempX], 0
	pusha
	invoke SetFilePointer, ebx, 0, tempX, FILE_CURRENT
	mov [x], eax
	popa
	upsh [x]
	upsh [tempX]
	xor ebx, ebx
	cmp [x], -1
	jne .tell.0
	inc ebx
.tell.0:
	upsh ebx
	ret
ENDP io_tell

_makeexe:	
	drop
	ret
;---------------------------------------------------------------------
; ANS MEMORY ALLOCATION WORDS
;---------------------------------------------------------------------

; ( u -- a-addr )

PROC mem_alloc
;	dup
	invoke LocalAlloc, 0, eax
alloc_return:
	mov ebx, eax
alloc_return2:
	cmp ebx, 1
	sbb ebx, ebx
	mov [__ior], ebx
	ret
ENDP mem_alloc

; ( a-addr -- )
PROC mem_free
	invoke LocalFree, eax
	upop ebx
	jmp alloc_return2
ENDP mem_free

; ( a-addr u -- a-addr2 )
PROC mem_realloc
	upop ebx	; EAX -> addr, EBX -> newsize
	invoke LocalReAlloc, eax, ebx, 0
	jmp alloc_return
ENDP mem_realloc
macro OSDICT {
}

data import

  library kernel,'KERNEL32.DLL'

  import kernel,\
	LocalAlloc, 'LocalAlloc',\
	LocalReAlloc, 'LocalReAlloc',\
	LocalFree, 'LocalFree',\
	GetModuleHandle,'GetModuleHandleA',\
	GetModuleFileName,'GetModuleFileNameA',\
	GetCommandLine, 'GetCommandLineA',\
	ExitProcess,'ExitProcess',\
	GetStdHandle, 'GetStdHandle',\
	WriteFile, 'WriteFile',\
	ReadFile, 'ReadFile',\
	SetConsoleTitle, 'SetConsoleTitleA',\
	CreateFile, 'CreateFileA',\
	 FlushFileBuffers, 'FlushFileBuffers',\
	 SetEndOfFile, 'SetEndOfFile',\
	 GetLastError,'GetLastError',\
	 SetFilePointer, 'SetFilePointer',\
	 GetFileSize, 'GetFileSize',\
	CloseHandle,'CloseHandle',\
	LoadLibrary,'LoadLibraryA',\
	FreeLibrary,'FreeLibrary',\
	GetProcAddress,'GetProcAddress', \
	GetEnvironmentStrings, 'GetEnvironmentStringsA'
end data

; General code goes here: ---------------------------------------------
include "revacore.asm"
