| MM-060505   Replacement for the needs.f file from Reva 6.0
| ------------------------------------------------------------------------------
|               Support to load Forth source code modules.
|
|                Copyright (C) 2006-2008 Manfred Mahlow
| 
| ------------------------------------------------------------------------------

| We support two module directories, so that we need two string buffers for
| the path names :

| cell | byte  | count chars   | byte |
| size | count | the path name | 0    |

  82 dup , here swap allot constant dir(1)  dir(1) off

  82 dup , here swap allot constant dir(2)  dir(2) off

| dir(x) returns the address of the count byte

loc:

 : .dirs ( -- )
 | Print the current module directories 
   ." dir(1) : " dir(1) count type cr ." dir(2) : " dir(2) count type ;

 variable msg  | controls the error message printing

 : ?msg ( a u -- )
   msg @ if type ." .4th " else 2drop then msg off
 ;

 create s  257 allot  | a buffer for the absolute module path name


0 [IF]
 : path ( a1 u1 a2 u2 -- a3 u3 )
 | Build and return the absolute path name a3,u3 from the module file name a1,u1
 | and the absolute module directory name a2,u2.
   dup if dup 3 pick +
          255 > if ." needs/take ?  path name overflow  " -1 throw then
          s place  'ps s c+place  s +place " .4th" s +place s count
       else 2drop
       then ;
[THEN]

 : path ( a1 u1 a2 u2 -- a3 u3 )
 | Build and return the absolute path name a3,u3 from the module file name a1,u1
 | and the absolute module directory name a2,u2.
   dup if dup 3 pick +
          255 > if ." needs/take ?  path name overflow  " -1 throw then
          s place  'ps s c+place  s +place
       else
         2drop s place
       then
       " .4th" s +placez s count ;

 : try ( a u dir -- 0|throw-code )
 | Try to include the module a,u from the module directory dir and return a
 | false flag on success or a throw-code on error.
   count path ['] included catch dup if nip nip then ;

 : (taken) ( a u -- a u 0|throw-code )
   | Try to load the module a,u from the module directories. The module name is
   | blank or 'pathsep' separated. Replace blanks with pathseps, append the 
   | module file name extension and try to include the file first from dir(1)
   | and then from dir(2). Return the module name a,u and zero on success or
   | the throw-code on error.
   dup >r over >r
   msg on  | error messages on
   verbose @ if 2dup cr ." loading " type ." .4th  " then
   | ** try to load the module from directory dir(1)
   dir(1) try ( -- f )
   ?dup if | ** loading from dir(1) failed
     dup -38 <>if r> r> ?msg exit then  | exit if no file I/O error
     drop  r> r> 2dup >r >r ( -- a u )  | on file I/O error, drop error code
     dir(2) try ( -- f )                | and try to load the module from dir(2)
     ?dup if r> r> ?msg exit then       | exit on error
   then
   r> r> ( -- a u ) verbose @ if        | display done message in verbose mode
     cr ."    done " 2dup type ." .4th  "
   then 0
 ;

| +++++++++++++++++ hier weiter +++++++++++++++++++++++++++++

 variable modules  modules off   | A wordlist to register loaded modules.

 : needed? ( a u -- a u tf|ff )
 | Return the module name a,u and a true flag, if the requested module a,u is
 | not yet loaded. Return a false flag otherwise.
   s.context push (esm) push  push-order  | push search context
   modules dup context ! s.context ! 0 context cell+ ! find not ;

 : register ( a u -- )
 | Register the module a,u as loaded.
   current push  modules current !  created ;

 : alloc ( a1 u -- a2 )
 | Allocate a buffer for the string a1,u , place the string zero-terminated in
 | that buffer, replace all blanks with the pathsep character and return the
 | buffer address on the stack.
 | dup 2 + allocate ioerr @ if ." memory allocation error " -1 throw then
   dup 2 + allocate if ." memory allocation error " -1 throw then
   dup >r place r> count bl2ps (zt) 1- ;

 : needed ( a u -- )
 | Load and register the module a,u if it's not already loaded.
   alloc dup >r count needed?
   if (taken) ?dup if r> free drop throw then register then r> free drop ;

 : module ( "ccc.."<  > -- a u )
 | Read the next phrase from the input stream and return the string parameters
 | on the stack. Abort with message if the input stream is exhausted.
   phrase dup 0if -31 throw then ;

 : needs ( "ccc.."<  > -- )
 | Read a module name from the input stream and load and register the module,
 | if it's not yet loaded. 
   module needed ;

 : taken ( a u -- )
 | Load the module a,u without registering it, so that it can be loaded again.
   alloc dup >r count (taken) ?dup if r> free drop throw then
   2drop r> free drop ;

 : take ( -- )
 | Read a module name from the input stream and load the module without 
 | registering it, so that it can be loaded again.
   module taken ;

 : .modules ( -- )
 | Print the absolute path of the module directories and all registered
 | modules.
   .dirs
   cr " ----------------------------------------" 2dup type type
   cr context @ modules dup @ 0if 2drop exit then  context ! words context ! ;

  reveal path
  reveal needed
  reveal needs
  reveal taken
  reveal take
  reveal .modules

  loc;

| ------------------------------------------------------------------------------
| MM-090809  (include) --> included
| MM-090809  errno -37 changed to -38
| MM-090305  +place -> +placez in path
| MM-090123  adapted to ANS conform allocate and free
| MM-080923  bug fixed in path
| MM-080920  (create) renamed to created
| MM-080916  (needs) -> needed , (take) -> taken , ((take)) -> (taken)
| MM-080914  Code partly rewritten to store module names no longer in the
|            rotating string buffer but in allocated memory
| MM-080322  .slibs renamed to .modules
| MM-080205  pathsep -> 'ps
| MM-071026  File name extension for module files changed from .rfl to .4th 
| MM-071019  ((take)) modfied not to save and restore the contexts any more.
|            constant core.wordlist removed
| MM-070516  debug -> verbose , error handling in ((take)) modified
| MM-070511  ;; -> exit
| MM-070310  File restructured, code partly rewritten, comments added.
| MM-070329  done message changed to not print a file name from the rotating
|            string buffer, if it was overwritten meanwhile, e.g. count <>
|            zcount.







