| MM-060505   Replacement for the needs.f file from Reva 6.0
| ------------------------------------------------------------------------------
| Support for forth source code library files.
|
| Copyright (C) 2006-2008 Manfred Mahlow
| 
| ------------------------------------------------------------------------------

| We support two library directories, so that we need two string buffers for
| the path names :

| cell | byte  | count chars   | byte |
| size | count | the path name | 0    |

  82 dup , here swap allot constant dir(1)  dir(1) off

  82 dup , here swap allot constant dir(2)  dir(2) off

| dir(x) returns the address of the count byte

loc:

 : .dirs ( -- )
 | Print the current values forth library dirs 
   ." dir(1) : " dir(1) count type cr ." dir(2) : " dir(2) count type ;


 variable msg  | to control the error message printing

 : .lib ( a u -- )
 | Print a library file name a,u and append the library file name extension.
   type ." .4th" ;

 : ?msg1 ( a u ? -- a u ? )
   >r msg @ if 2dup .lib ."   " then msg off r>
 ;

 : ?msg2 ( a u ? -- a u ? )
   dup if | an error occured
     ?msg1
   then
 ;

 create s  257 allot  | a string buffer to build the library path name

 : path ( a1 u1 a2 u2 -- a3 u3 )
 | Build and return the absolute path name a3,u3 from the library file name
 | a1,u1 and the directory path name a2,u2.
   dup if dup 3 pick +
          255 > if ." needs/take ?  path name overflow  " -1 throw then
          s place  'ps s c+place  s +place " .4th" s +place s count
       else 2drop
       then ;

 : try ( a u dir -- a u f )
 | Try to include the library file a,u from the library directory dir(x) and
 | return the file name a,u and a true flag on success or a false flag on error.
   over >r 2 pick >r | save a,u
   count path ['] (include) catch dup if nip nip then r> r> rot ;
  | count path cr ." 0:" .s ['] (include) catch cr ." 1:" .s dup if nip nip then
  | r> r> 2dup type rot cr ." 2:" .s cr ;

 : ((take)) ( a u -- )
   | Try to load the library named a,u. The library name is blank or 'pathsep'
   | separated. Replace blanks with pathseps, append the library file name 
   | extension and try to include the file first from dir(1) and then from
   | dir(2). Abort with an error message if the file was not found or if an 
   | error occures while loading the library. 
   msg on  | error messages on
   verbose @ if 2dup cr ." loading " type ." .4th  " then
   | ** try to load the library from library directory dir(1)
   dir(1) try ( -- a u f )
   ?dup if | ** loading library from dir(1) failed
     dup -37 <>if ?msg1 throw then  | throw if no file I/O error
     drop  | ** on file I/O error, drop error code and try to load the
     dir(2) try ( -- a.l u.l f )  | library from library directory dir(2)
     ?msg2 throw
   then ( -- a u )
   verbose @ if  | ** display done message
     cr ."    done "
     >r zcount dup r> =  if type ." .4th  " else 2drop then
   else          | ** no done message
     2drop
   then
 ;

 variable libs  libs off   | A hidden wordlist. Libraries are registered here.

 : needed? ( a u -- a u f )
 | Return a true flag, if the requested library a,u must be loaded, because 
 | that was not yet done. Return a false flag otherwise.
   s.context push (esm) push  push-order  | push search context
   libs dup context ! s.context ! 0 context cell+ ! find not ;

 : register ( a u -- )
 | Register the library a,u as loaded.
   current push  libs current !  (create) ;

 : file ( a1 u -- a2 u )
 | Copy the string a1,u to the temporary string buffer, replace all blanks with
 | pathseps and return the updated string parameters a2,u on the stack.
   "" bl2ps ;

 : (needs) ( a u -- )
 | Load and register the library a,u if it's not already loaded.
   file needed? if 2dup >r >r ((take)) r> r> register then ;

 : library ( "ccc.."<  > -- a u )
 | Read the next phrase from the input stream and return the string parameters
 | on the stack. Abort with message if the input stream is exhausted.
   phrase dup 0if -31 throw then ;

 : needs ( "ccc.."<  > -- )
 | Read a library name from the input stream and load and register the library,
 | if it's not yet loaded. 
   library (needs) ;

 : (take) ( a u -- )
 | Load the library a,u without registering it, so that it can be loaded again.
   file ((take)) ;

 : take ( -- )
 | Read a library name from the input stream and load the library without 
 | registering it, so that it can be loaded again.
   library (take) ;

 : .modules ( -- )
 | Print the absolute path of the library directories and all registered
 | libraries.
   .dirs
   cr " ----------------------------------------" 2dup type type
   cr context @ libs dup @ 0if 2drop exit then  context ! words context ! ;

  reveal (needs)
  reveal needs
  reveal (take)
  reveal take
  reveal .slibs
  reveal .modules

  loc;

| ------------------------------------------------------------------------------
| MM-080914  "reveal .slibs" removed, that was overseen om 080322
| MM-080322  .slibs renamed to .modules
| MM-070310  File restructured, code partly rewritten, comments added.
| MM-070329  done message changed to not print a file name from the rotating
|            string buffer, if it was overwritten meanwhile, e.g. count <> zcount.
| MM-070511  ;; -> exit
| MM-070516  debug -> verbose , error handling in ((take)) modified
| MM-071019  ((take)) modfied not to save and restore the contexts any more.
|            constant core.wordlist removed
| MM-071026  File name extension for library files changed from .rfl to .4th 
| MM-080205  pathsep -> 'ps

